# Copyright (C)  2024  Robert Labudda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Data model classes"""
import enum
import datetime
import tomllib
import logging
from pathlib import Path
from dataclasses import dataclass

from nttl.dt import dt_as_str


TOML_TABLE_DT_FORMAT = "%Y-%m-%dT%H-%M-%S"


class Action(enum.StrEnum):
    START = 'start'
    STOP = 'stop'
    SWITCH = 'switch'
    EVENT = 'event'


@dataclass
class Entry:
    timestamp: datetime.datetime
    action: str
    description: str = ''
    labels: list[str]|None = None
    startentry: datetime.datetime|None = None

    def __post_init__(self):
        if self.labels is None:
            self.labels = []

    def as_toml(self):
        lines = ['[' + dt_as_toml_table_str(self.timestamp) + ']',
                 'action = "' + self.action + '"']

        if len(self.description) > 0:
            text = self.description
            wrap = '"'
            if "\n" in text:
                wrap = "'''"
            elif "'" not in text:
                wrap = "'"
            elif '"' in text:
                text = text.replace('"', '\\"')
            lines += ['description = ' + wrap + text + wrap]

        assert self.labels is not None
        labels = {label.strip()
                  for label in self.labels
                  if label is not None and len(label.strip()) > 0}
        if len(labels) > 0:
            lines += ['labels = ' + repr(sorted(list(labels)))]

        if self.action == Action.STOP and self.startentry is not None:
            lines += ['startentry = ' + dt_as_str(self.startentry)]
        return "\n".join(lines) + "\n\n"

    def __lt__(self, other):
        return self.timestamp < other.timestamp


def dt_as_toml_table_str(value):
    return value.strftime(TOML_TABLE_DT_FORMAT)


def toml_table_str_as_dt(value):
    return datetime.datetime.strptime(value, TOML_TABLE_DT_FORMAT)


class Activity:
    """The time span of a task"""
    def __init__(self, start, end):
        self.start = start
        self.end = end
        if end < start:
            self.start, self.end = end, start

    def get_labels(self):
        return list(sorted(set(self.start.labels) | set(self.end.labels)))

    def duration(self):
        return self.end.timestamp - self.start.timestamp

    def __contains__(self, dt):
        """Return True if dt is between start and end"""
        return self.start.timestamp <= dt <= self.end.timestamp

    def description(self):
        return ' '.join([a.description
                         for a in [self.start, self.end]
                         if len(a.description) > 0])

    def __lt__(self, other):
        return self.start < other.start

    def overlaps(self, start, end):
        if isinstance(start, Entry):
            start = start.timestamp
        if isinstance(end, Entry):
            end = end.timestamp
        if end < start:
            end, start = start, end
        mystart = self.start.timestamp
        myend = self.end.timestamp
        return any([
            start >= mystart and end <= myend,
            mystart >= start and myend <= end,
            start >= mystart and start <= myend,
            mystart >= start and mystart <= end,
            end <= myend and end >= mystart,
            myend <= end and myend >= start,
            ])


def load_entries(datafile):
    datafile = Path(datafile)
    if not datafile.is_file():
        return

    with open(datafile, "rb") as datafile:
        raw = tomllib.load(datafile)

    for key, data in raw.items():
        try:
            timestamp = toml_table_str_as_dt(key)
        except ValueError:
            logging.warning(f"Not a timestamp: {key}")
            continue
        entry = Entry(timestamp,
                      data['action'],
                      data.get('description', ''),
                      data.get('labels', None),
                      data.get('startentry', None))
        yield entry


def get_entries_between(datafile, start, end):
    if not isinstance(start, datetime.datetime):
        assert isinstance(start, datetime.date)
        start = datetime.datetime(start.year,
                                  start.month,
                                  start.day,
                                  0, 0, 0)
    if not isinstance(end, datetime.datetime):
        assert isinstance(end, datetime.date)
        end = datetime.datetime(end.year,
                                end.month,
                                end.day,
                                23, 59, 59)

    for entry in sorted(load_entries(datafile)):
        if entry.timestamp < start:
            continue
        if entry.timestamp > end:
            break

        yield entry



# Copyright (C)  2024  Robert Labudda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Datetime/str utilities"""
import datetime
import locale


WEEKDAYS = [locale.nl_langinfo(getattr(locale, f'DAY_{n}')).lower()
            for n in range(1, 7)]
SHORT_WEEKDAYS = [locale.nl_langinfo(getattr(locale, f'ABDAY_{n}')).lower()
                  for n in range(1, 7)]
C_WEEKDAYS = ['monday', 'tuesday', 'wednesday', 'friday',
              'saturday', 'sunday']
C_SHORT_WEEKDAYS = ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun']
ALL_WEEKDAYS = set(WEEKDAYS + SHORT_WEEKDAYS + C_WEEKDAYS + C_SHORT_WEEKDAYS)


def parse_when(value):
    now = datetime.datetime.now()
    if value is None or len(value.strip()) == 0:
        return now

    value = value.strip()

    if value[0] == 'T':
        return now + parse_timedelta(value[1:])
    elif value[0] in '-+':
        return now + parse_timedelta(value)

    return parse_absolute_dt(value)


def parse_absolute_dt(value):
    now = datetime.datetime.now()
    if len(value) == 5:
        return datetime.datetime.strptime(value,
                                          '%H:%M').replace(year=now.year,
                                                           month=now.month,
                                                           day=now.day)
    elif len(value) == 8:
        return datetime.datetime.strptime(value,
                                          '%H:%M:%S').replace(year=now.year,
                                                              month=now.month,
                                                              day=now.day)
    elif len(value) == 17:
        return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M')
    elif len(value) == 19:
        return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')

    raise ValueError(f"Don't know the time format '{value}'")


def parse_absolute_date_or_dt(value):
    if len(value) == 10:
        return datetime.datetime.strptime(value, '%Y-%m-%d')
    return parse_absolute_dt(value)


def dt_as_str(value):
    return value.strftime("%Y-%m-%dT%H:%M:%S")


def parse_dt(value):
    try:
        return parse_absolute_date_or_dt(value)
    except ValueError:
        pass

    now = datetime.datetime.now()

    if value.lower() == 'today':
        return now.date()

    if value.lower() in WEEKDAYS + SHORT_WEEKDAYS:
        value = value.lower()

        if value in WEEKDAYS:
            fmt = '%A'
        else:
            fmt = '%a'

        then = now - datetime.timedelta(days=1)
        # walk backwards day by day until the locale of the weekday
        # matches what was asked for
        while then.strftime(fmt).lower() != value:
            then -= datetime.timedelta(days=1)

        return then.date()

    if value.lower() in C_WEEKDAYS + C_SHORT_WEEKDAYS:
        value = value.lower()

        if value in C_WEEKDAYS:
            idx = C_WEEKDAYS.index(value)
        else:
            idx = C_SHORT_WEEKDAYS.index(value)

        days = (now.weekday() - idx)%7
        if days == 0:
            days = 7
        then = now - datetime.timedelta(days=days)
        return then.date()

    return (now + parse_timedelta(value)).date()


def parse_timedelta(text):
    if text[0] == 'T':
        text = text[1:]

    sign = -1 if text[0] == '-' else 1
    if text[0] in '-+':
        text = text[1:]

    value = ''
    duration = datetime.timedelta(0)
    for letter in text:
        if letter == '_' or letter.isspace():
            continue
        if letter.isdigit():
            value += letter
        if letter in 'hmsdw':
            if len(value) == 0:
                continue

            if letter == 'h':
                duration += datetime.timedelta(hours=int(value))
            elif letter == 'm':
                duration += datetime.timedelta(minutes=int(value))
            elif letter == 's':
                duration += datetime.timedelta(seconds=int(value))
            elif letter == 'd':
                duration += datetime.timedelta(days=int(value))
            elif letter == 'w':
                duration += datetime.timedelta(days=7*int(value))

            value = ''

    if len(value) > 0:
        duration += datetime.timedelta(minutes=int(value))

    return sign * duration



# Copyright (C)  2024  Robert Labudda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""nttl program module"""
import sys
import argparse
import datetime
from pathlib import Path

from nttl import version
from nttl.dt import parse_when, parse_dt
from nttl.data import Action, Entry, load_entries, get_entries_between
from nttl.report import print_report

try:
    from xdg import BaseDirectory
except ImportError:
    BaseDirectory = None


HERE = Path(__file__).absolute().parent
HOME = Path().home()
PROGRAMNAME = 'nttl'
CONFIGDIR = HOME / ".config" / PROGRAMNAME
DATADIR = HOME / ".local" / "share" / PROGRAMNAME

if BaseDirectory is not None:
    CONFIGDIR = Path(BaseDirectory.save_config_path(PROGRAMNAME) or CONFIGDIR)
    DATADIR = Path(BaseDirectory.save_data_path(PROGRAMNAME) or DATADIR)

CONFIGFILE = CONFIGDIR / "config.toml"
DATAFILE = DATADIR / "times.toml"


def error(text, *args, **kwargs):
    print(text, *args, file=sys.stderr, **kwargs)


def get_started_tasks(datafile):
    now = datetime.datetime.now()
    active_tracks = {}
    for entry in sorted(load_entries(datafile)):
        if entry.timestamp > now:
            continue
        if entry.action == Action.START:
            active_tracks[entry.timestamp] = entry
        if entry.action == Action.SWITCH:
            active_tracks = {entry.timestamp: entry}
        if entry.action == Action.STOP:
            if entry.startentry is None:
                active_tracks = {}
                continue
            if entry.startentry not in active_tracks:
                continue
            del active_tracks[entry.startentry]

    for entry in sorted(active_tracks.values()):
        if entry.timestamp > now:
            break
        yield entry


def add_entry_params(parser):
    parser.add_argument('-d', '--description',
                        default='',
                        type=str)
    parser.add_argument('-l', '--label',
                        default=[],
                        action="append",
                        type=str)
    add_when_parameter(parser)


def add_when_parameter(parser):
    parser.add_argument('-w', '--when',
                        type=str,
                        default=None,
                        help="Set the time when this happened. Either as "
                             "absolute time, date+time, or relative. "
                             "A date+time is written as 'YYYY-MM-DD HH:MM:SS', "
                             "time is written as 'HH:MM:SS'. The seconds are "
                             "optional. Relative time is written 'T-2h5m10s'. "
                             "'h', 'm', and 's' can be mixed and combined. If "
                             "you only provide a number (e.g. 'T-5'), 'm' is "
                             "assumed.")


def add_start_end_parameters(parser):
    parser.add_argument('-s', '--start',
                        type=str,
                        default="-1w",
                        help="Start of report period. Relative date, "
                             "absolute date, or day of week in your locale. "
                             "Default is '-1w'.")
    parser.add_argument('-e', '--end',
                        type=str,
                        default="0d",
                        help="End of report period. Relative date, absolute "
                             "date, or day of week in your locale. Default is "
                             "'today'.")


def add_task_id_parameter(parser):
    parser.add_argument('-t', '--task',
                        type=int,
                        default=None,
                        help="Active task id. If not provided, "
                             "all active tasks are stopped.")


def print_status(sourcefile, outputformat, customformat):
    if outputformat == 'text':
        print("Active tasks:")
    elif outputformat == 'json':
        print('[')

    started_tasks = list(get_started_tasks(sourcefile))
    for idx, entry in enumerate(started_tasks):
        assert isinstance(entry.labels, list)
        if outputformat == 'text':
            print(f"({idx+1})",
                  entry.timestamp,
                  entry.description,
                  entry.labels)
        elif outputformat == 'json':
            comma = ','
            if idx+1 >= len(started_tasks):
                comma = ''
            labels = ', '.join('"' + l + '"' for l in entry.labels)
            print(f''' {{"start": "{entry.timestamp}",
"description": "{entry.description}",
"labels": [{labels}]
}}{comma}''')
        elif outputformat == 'csv':
            labels = ",".join(l for l in entry.labels)
            print(f'{entry.timestamp},"{entry.description}","{labels}"')

        elif outputformat == 'custom':
            labels = ', '.join(l for l in entry.labels)
            descr_or_labels = entry.description
            if len(descr_or_labels.strip()) == 0:
                descr_or_labels = labels
            print(customformat.format(timestamp=str(entry.timestamp),
                                      description=str(entry.description),
                                      descr_or_labels=descr_or_labels,
                                      labels=labels))

    if outputformat == 'json':
        print(']')


def run():
    CONFIGDIR.mkdir(exist_ok=True, parents=True)
    DATADIR.mkdir(exist_ok=True, parents=True)

    parser = argparse.ArgumentParser(prog=PROGRAMNAME)
    parser.add_argument('-v', '--version',
                        action='version',
                        version='%(prog)s ' + version.__version__)
    subparsers = parser.add_subparsers(dest="fn")

    startparser = subparsers.add_parser(Action.START,
                                        help="Start another task.")
    add_entry_params(startparser)

    stopparser = subparsers.add_parser(Action.STOP,
                                       help="Stop an active task.")
    add_entry_params(stopparser)
    add_task_id_parameter(stopparser)

    switchparser = subparsers.add_parser(Action.SWITCH,
                                         help="Switch tasks (stop active, "
                                              "start another)")
    add_entry_params(switchparser)
    add_task_id_parameter(switchparser)

    eventparser = subparsers.add_parser(Action.EVENT,
                                        help="Add an event")
    add_entry_params(eventparser)

    statusparser = subparsers.add_parser('status', help="Show the active tasks")
    statusparser.add_argument('-f', '--format',
                              choices=['text', 'json', 'csv', 'custom'],
                              default='text',
                              help="Output format. 'json', 'csv', 'custom', or "
                                   "'text' (the default).")
    statusparser.add_argument('-c', '--custom',
                              default="{timestamp}: {description} ({labels})",
                              help="The format of the custom format.")

    listparser = subparsers.add_parser('list', help="List all task tracking events")
    add_start_end_parameters(listparser)

    reportparser = subparsers.add_parser('report', help="Show a report about time spent")
    add_start_end_parameters(reportparser)
    reportparser.add_argument('-d', '--description',
                              type=str,
                              default=[],
                              action='append',
                              help="Filter by description. Each -d phrase must "
                                   "be in a tasks description to be included.")
    reportparser.add_argument('-l', '--label',
                              default=[],
                              action='append',
                              help='Filter by label. Each -l label must appear '
                                   'in a task to be included')
    reportparser.add_argument('-x', '--exclude-label',
                              default=[],
                              action='append',
                              help='Exclude activities that have this label. '
                                   'May be provided multiple times.')
    reportparser.add_argument('-f', '--format',
                              choices=['csv', 'text', 'bar', 'gantt'],
                              default='text',
                              help='What format to use. Options are csv or '
                                   'text. "text" is the default.')
    reportparser.add_argument('-g', '--group-by-labels',
                              default=False,
                              action='store_true',
                              help='Group by labels')
    reportparser.add_argument('--resolution',
                              default='1h',
                              help='Resolution for graphs. Default is '
                                   '"%(default)s".')

    args = parser.parse_args()
    if args.fn is None:
        parser.print_help()
        return

    if args.fn in list(Action):
        entry = Entry(parse_when(args.when),
                      args.fn)
        entry.description = args.description
        if args.label is not None:
            entry.labels = args.label

        if args.fn in [Action.STOP, Action.SWITCH]:
            if args.task is not None:
                active_tasks = list(get_started_tasks(DATAFILE))
                entry.startentry = active_tasks[args.task-1].timestamp

        with open(DATAFILE, "a+t", encoding='utf-8') as datafile:
            datafile.write(entry.as_toml())

    if args.fn == 'list':
        time_range = [parse_dt(args.start),
                      parse_dt(args.end)]

        for entry in sorted(get_entries_between(DATAFILE, *time_range)):
            print(entry.timestamp,
                  f"({entry.action})",
                  entry.description,
                  entry.labels)

    if args.fn == 'status':
        print_status(DATAFILE, args.format, args.custom)

    if args.fn == 'report':
        print_report(DATAFILE, args, error=error)



# Copyright (C)  2024  Robert Labudda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Time reporting"""
import sys
import csv
import datetime

from nttl.data import Action, Activity, Entry, get_entries_between
from nttl.dt import parse_dt, parse_timedelta


BARS = ' ▂▄▆█'
BAR_WIDTH = 4
SLICES = '▏▎▍▌▋▊▉█'


def ignore_error(*_):
    # dummy function to ignore errors
    pass


def delta_as_str(delta):
    seconds = delta.seconds
    hours = delta.days*24 + seconds//3600
    seconds %= 3600

    minutes = seconds//60
    seconds %= 60

    return f"{hours}:{minutes:0>2}:{seconds:0>2}"


def print_report(sourcefile, args, error=None):
    if error is None:
        error = ignore_error
    now = datetime.datetime.now()
    start = parse_dt(args.start)
    end = parse_dt(args.end)

    started_tasks = {}
    activities = []
    all_labels = set()

    # collect all start/(stop|switch) pairs
    for entry in get_entries_between(sourcefile, start, end):
        if entry.labels is not None:
            all_labels.update(set(entry.labels))

        if entry.action == Action.STOP:
            if entry.startentry is None:
                for started in started_tasks.values():
                    activities.append(Activity(started, entry))
                started_tasks = {}
                continue

            if entry.startentry not in started_tasks:
                error("Task ended but did not start in report period")
                continue

            activities.append(Activity(started_tasks[entry.startentry], entry))
            del started_tasks[entry.startentry]

        if entry.action == Action.START:
            started_tasks[entry.timestamp] = entry

        if entry.action == Action.SWITCH:
            for started in started_tasks.values():
                activities.append(Activity(started,
                                           Entry(entry.timestamp, Action.STOP)))
            started_tasks = {entry.timestamp: entry}

    # some tasks are still running? pretend they're done now
    for entry in started_tasks.values():
        activities.append(Activity(entry, Entry(now, Action.STOP)))

    # accumulate by labels
    by_labels = {label: [] for label in all_labels}

    filtered_sorted = []
    for activity in activities:
        labels = activity.get_labels()
        if len(args.label) > 0 and not all(l in labels for l in args.label):
            continue
        if len(args.description) > 0 and not all(d in activity.start.description for d in args.description):
            continue
        # accumulate by labels
        for label in activity.get_labels():
            by_labels[label].append(activity)

        filtered_sorted.append(activity)
    filtered_sorted.sort()

    # don't show labels without any registered activity
    by_labels = {lbl: act
                 for lbl, act in by_labels.items()
                 if len(act) > 0}
    for lbl in args.exclude_label:
        if lbl not in by_labels:
            continue
        del by_labels[lbl]

    if args.group_by_labels:
        lines = [[label, sum([a.duration() for a in activities],
                             start=datetime.timedelta(0))]
                  for label, activities in by_labels.items()]
        lines.sort()
    else:
        lines = [[a.start.timestamp,
                  a.end.timestamp,
                  a.duration(),
                  a.description(),
                  a.get_labels()] for a in filtered_sorted]

    if args.format == 'csv':
        writer = csv.writer(sys.stdout)
        if args.group_by_labels:
            for line in lines:
                writer.writerow([line[0], delta_as_str(line[1])])
        else:
            for line in lines:
                line[-1] = ','.join(line[-1])
                writer.writerow(line)

    elif args.format == 'bar':
        if len(activities) == 0:
            return
        labels = [lbl for lbl in sorted(by_labels.keys())]
        spacing = [max(len(lbl), BAR_WIDTH) for lbl in labels]
        durations = [int(sum([a.duration() for a in activities],
                             start=datetime.timedelta(0)).total_seconds())
                     for _, activities in sorted(by_labels.items())]
        # durations is in seconds per label
        # convert to (full hours, remaining minutes)
        durations = [(d // 3600,  (d % 3600)//60)
                     for d in durations]
        max_hours = max(d[0] for d in durations) + 1
        for line in range(0, max_hours+1):
            hour = max_hours - line
            print(f" {hour: >3}  │ ", end="") 
            for idx, label in enumerate(labels):
                pad = spacing[idx]
                if durations[idx][0] >= hour:
                    symbol = BARS[4]
                    if durations[idx][0] == hour:
                        symbol = BARS[int(4*durations[idx][1]/60.0)]
                    pad = (pad-BAR_WIDTH)//2
                    print(' '*(pad+1) + symbol*BAR_WIDTH + ' '*(pad+1), end="")
                    if spacing[idx] % 2 == 1:
                        print(' ', end="")
                else:
                    print(' '*(pad + 2), end="")
            print()

        # x-axis
        print("──────┼─" + "─"*(sum([max(BAR_WIDTH, len(l))+2 for l in labels])))
        print("      │ ", end="")
        for idx, label in enumerate(labels):
            pad = max(0, BAR_WIDTH-len(label))
            print(" "*max(1, pad) + label + " "*max(1, pad), end="")
        print()

    elif args.format == 'gantt':
        if len(activities) == 0:
            return
        start = min(a.start.timestamp for a in activities)
        end = max(a.end.timestamp for a in activities)
        granularity = parse_timedelta(args.resolution)
        ticks = int((end - start).total_seconds() // granularity.total_seconds()) + 2
        max_lbl_len = max(len(lbl) for lbl in by_labels.keys())+1
        for label, activities in sorted(by_labels.items()):
            print(" " + label, end="")
            print(" "*max(1, max_lbl_len-len(label)) + "│", end="")
            dt = start
            last_dt = start
            while dt < end:
                is_active = any(a.overlaps(last_dt, dt) for a in activities)
                symbol = " "
                if is_active:
                    symbol = "█"
                print(symbol, end="")
                last_dt = dt
                dt += granularity
            print()

        print("─"*(max_lbl_len+1) + "┼", end="")
        print("─"*(ticks))

        # hours x-axis
        print(" "*(max_lbl_len+1) + "│", end="")
        dt = start
        counter = 0
        while dt < end:
            if counter % 8 == 0:
                print("└ " + dt.strftime("%H:%M"), end=" ")
            dt += granularity
            counter += 1
        print()

        # dates x-axis
        print(" "*(max_lbl_len+1) + "│", end="")
        dt = start
        overshoot = 0
        last_date = None
        while dt < end:
            if dt.hour == 0 and last_date != dt.date():
                last_date = dt.date()
                text = "└ " + dt.strftime("%Y-%m-%d") + " "
                overshoot = len(text)
                print(text, end="")
            elif overshoot > 0:
                overshoot -= 1
            else:
                print(" ", end="")
            dt += granularity

        print()

    else:
        if args.group_by_labels:
            start = filtered_sorted[0].start.timestamp.strftime("%x")
            end = filtered_sorted[-1].end.timestamp.strftime("%x")
            print(f"{start} --- {end}")
            for line in lines:
                print(f"{line[0]}    {line[1]}")
        else:
            day = None
            for line in lines:
                if line[0].day != day:
                    day = line[0].day
                    print(line[0].strftime("%x"))
                start = line[0].strftime("%H:%M")
                end = line[1].strftime("%H:%M")
                labels = ", ".join(line[-1])
                description = line[3]
                print(f"  {start} - {end}    {labels}    {description}")
